/*
 * Copyright (C) 2005 Andre Noll <noll@mathematik.tu-darmstadt.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "audiod.cmdline.h"
#include "para.h"

#define MAX_STAT_CLIENTS 50
#define MAX_PENALTY 20
#define MAX_PENALTY_MS 2000
#define PENALTY_RESET_TIME 10
#define PENALTY_INCREASE_TIME 2

extern const char *status_item_list[NUM_STAT_ITEMS];
static int stat_pipe = -1, signal_pipe[2];
struct gengetopt_args_info conf;
static int offset_seconds, length_seconds, sa_time_diff_sign = 1;
static struct timeval server_stream_start, sa_time_diff;
static int current_decoder = -1, num_stat_clients;

struct audio_format_info{
	char *name;
	char *rcmd, *wcmd;
	int bufsize;
	unsigned long penalty;
	/* last start time of the decoder for that format in any slot */
	struct timeval stime;
};


static struct audio_format_info afi[] = { /* audio format info */
#if 1
	{
		.name = "mp3",
		.rcmd = "para_recv",
		.wcmd = "madplay -",
		.bufsize = 4096,
	},
#endif
#ifdef HAVE_OGGVORBIS
	{
		.name = "ogg",
		.rcmd = "para_recv --header",
		.wcmd = "ogg123 -",
		.bufsize = 4096,
	},
#endif
	{
		.name = NULL,
	}
};

struct stream_io {
	int format, read_fd, write_fd;
	pid_t rpid, wpid;
	/* last read/write time, last start time od reader/writer */
	struct timeval rtime, wtime, rstime, wstime;
	int buflen;
	char *buf;
	int rkilled, wkilled;
};

struct command {
	char *name;
	int (*handler)(int, int, char**);
	char *description;
	char *usage;
	char *help;
};
int com_cycle(int, int, char **);
int com_help(int, int, char **);
int com_off(int, int, char **);
int com_on(int, int, char **);
int com_sb(int, int, char **);
int com_stat(int, int, char **);
int com_term(int, int, char **);
struct command cmds[] = {
	{
		.name = "cycle",
		.handler = com_cycle,
		.description = "switch to next mode",
		.usage = "cycle",
		.help = "on -> standby -> off -> on\n"
	},
	{
		.name = "help",
		.handler = com_help,
		.description = "display command list or help for given command",
		.usage = "help [command]",
		.help =
                "When I was younger, so much younger than today, I   \n"
                "never needed anybody's help in any way.  But now    \n"
                "these days are gone, I'm not so self assured. Now I \n"
                "find I've changed my mind and opened up the doors.  \n"
                "-- Beatles: Help                                    \n"
	},
	{
		.name = "off",
		.handler = com_off,
		.description = "deactivate para_audiod",
		.usage = "off",
		.help =
                "Close connection to para_server and stop all        \n"
                "decoders.                                           \n"
	},
	{
		.name = "on",
		.handler = com_on,
		.description = "activate para_audiod",
		.usage = "on",
		.help =
                "Establish connection to para_server, retrieve       \n"
                "para_server's current status. If playing, start     \n"
                "corresponding decoder. Otherwise stop all decoders. \n"
	},
	{
		.name = "sb",
		.handler = com_sb,
		.description = "enter standby mode",
		.usage = "sb",
		.help =
                "Stop all decoders but leave connection to           \n"
                "para_server open.                                   \n"
	},
	{
		.name = "stat",
		.handler = com_stat,
		.description = "print status information",
		.usage = "stat",
		.help =
		"Add para_audiod status information to para_server's \n"
		"status information and dump everything to stdout.\n"
	},
	{
		.name = "term",
		.handler = com_term,
		.description = "terminate audiod",
		.usage = "term",
		.help =
                "Stop all decoders, shut down connection to          \n"
                "para_server and exit.                               \n"
	},
	{
		.name = NULL,
	}
};

#define MAX_STREAM_SLOTS 5
static struct stream_io siol[MAX_STREAM_SLOTS];
static int playing;
static char *af_status; /* the audio format announced in server status */
static FILE *logfile;
static int audiod_socket = -1;
static struct sockaddr_un unix_addr;
static char *socket_name, *hostname;

#define AUDIOD_OFF 1
#define AUDIOD_ON 2
#define AUDIOD_STANDBY 3

static int audiod_status = AUDIOD_ON;

/*
 * log function. first argument is loglevel.
 */
void clog(int ll, char* fmt,...)
{
	va_list argp;
	FILE *outfd;
	struct tm *tm;
	time_t t1;
	char str[MAXLINE] = "";

	if (ll < conf.loglevel_arg)
		return;
	if (!logfile && conf.logfile_given)
		logfile = open_log(conf.logfile_arg);
	if (!logfile && conf.daemon_given)
		return;
	if (!logfile) {
		if (ll < WARNING)
			outfd = stdout;
		else
			outfd = stderr;
	} else
		outfd = logfile;
	time(&t1);
	tm = localtime(&t1);
	strftime(str, MAXLINE, "%b %d %H:%M:%S", tm);
	fprintf(outfd, "%s %s ", str, hostname);
	if (conf.loglevel_arg <= VERBOSE)
		fprintf(outfd, "%i ", ll);
	va_start(argp, fmt);
//	if (conf.loglevel_arg > VERBOSE)
//		dummy = va_arg(argp, char*);
	vfprintf(outfd, fmt, argp);
	va_end(argp);
}

static char **ring_buffer;
static int rb_size = 30;
static int rb_bytes;
long long unsigned rb_serial;

static void init_ring_buffer(void)
{
	ring_buffer = para_malloc(rb_size * sizeof(char *));
	memset(ring_buffer, 0, rb_size * sizeof(char*));
}

static void ring_buffer_free(void)
{
	int i;
	for (i = 0; i < rb_size; i++) {
		if (ring_buffer[i]) {
			free(ring_buffer[i]);
			ring_buffer[i] = NULL;
		}
	}
}

static void ring_buffer_add(char *line)
{
	unsigned int p = rb_serial % rb_size;

	if (!line)
		return;
	if (ring_buffer[p]) {
		rb_bytes -= strlen(ring_buffer[p]);
		free(ring_buffer[p]);
	}
	ring_buffer[p] = para_strdup(line);
	rb_bytes += strlen(line);
	rb_serial++;
}

char *get_time_string(struct timeval *newest_stime)
{
	struct timeval now, diff, adj_stream_start, tmp;
	int total = 0, use_server_time = 1;

	if (!playing)
		return para_strdup("");
	if (audiod_status == AUDIOD_OFF)
		goto out;
	if (sa_time_diff_sign > 0)
		tv_diff(&server_stream_start, &sa_time_diff,
			&adj_stream_start);
	else
		tv_add(&server_stream_start, &sa_time_diff,
			&adj_stream_start);
	tmp = adj_stream_start;
	if (newest_stime && audiod_status == AUDIOD_ON) {
		tv_diff(newest_stime, &adj_stream_start, &diff);
		if (tv2ms(&diff) < 5000) {
			tmp = *newest_stime;
			use_server_time = 0;
		}
	}
	gettimeofday(&now, NULL);
	tv_diff(&now, &tmp, &diff);
	total = diff.tv_sec + offset_seconds;
	if (total > length_seconds)
		total = length_seconds;
	if (total < 0)
		total = 0;
out:
	return make_message(
		"%s:%s%d:%02d [%d:%02d] (%d%%/%d:%02d)",
		status_item_list[SI_PLAY_TIME],
		use_server_time? "~" : "",
		total / 60,
		total % 60,
		(length_seconds - total) / 60,
		(length_seconds - total) % 60,
		length_seconds? (total * 100 + length_seconds / 2) / length_seconds : 0,
		length_seconds / 60,
		length_seconds % 60
	);
}


static char *audiod_status_string(void)
{
	int i;
	struct timeval *newest_stime = NULL;
	char *ret, *time_string, *uptime_string, *decoder_flags =
		para_malloc((MAX_STREAM_SLOTS + 1) * sizeof(char));

	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		struct stream_io *s = &siol[i];
		char flag = '0';
		if (s->rpid > 0)
			flag += 1;
		if (s->wpid > 0)
			flag += 2;
		if (flag != '0')
			flag += s->format * 4;
		decoder_flags[i] = flag;
		if (s->wpid <= 0)
			continue;
		if (newest_stime && tv_diff(&s->wstime, newest_stime, NULL) <= 0)
			continue;
		newest_stime = &s->wstime;
	}
	decoder_flags[MAX_STREAM_SLOTS] = '\0';
	time_string = get_time_string(newest_stime);
	uptime_string = uptime_str(server_uptime(GET));
	ret = make_message("%s:%s\n%s:%s\n%s:%s\n%s",
		status_item_list[SI_AUDIOD_UPTIME], uptime_string,
		status_item_list[SI_DECODER_FLAGS], decoder_flags,
		status_item_list[SI_AUDIOD_STATUS], audiod_status == AUDIOD_ON?
			"on" : (audiod_status == AUDIOD_OFF? "off": "sb"),
		 time_string);
	free(uptime_string);
	free(decoder_flags);
	free(time_string);
	return ret;
}

static int ring_buffer_dump(int fd)
{
	unsigned int p = rb_serial % rb_size;
	char *tmp, *buf = para_strdup(NULL);
	int ret;
	ssize_t len;

	do {
		if (ring_buffer[p]) {
			tmp = buf;
			buf = make_message("%s%s\n", tmp, ring_buffer[p]);
			free(tmp);
		}
		p = (p + 1) % rb_size;
	} while (p != (rb_serial % rb_size));
	tmp = audiod_status_string();
	buf = para_strcat(buf, tmp);
	buf = para_strcat(buf, "\n");
	free(tmp);
	len = strlen(buf);
	ret = send_buffer(fd, buf);
//	clog(DEBUG, "%s: wrote %d/%d\n", __func__, ret, len);
	free(buf);
	return ret;
}

static char *configfile_exists(void)
{
	static char *config_file;

	if (!config_file) {
		char *home = para_homedir();
		config_file = make_message("%s/.paraslash/audiod.conf", home);
		free(home);
	}
	return file_exists(config_file)? config_file : NULL;
}

static void setup_signal_handling(void)
{
	if (para_signal_init(signal_pipe, clog) < 0)
		exit(EXIT_FAILURE);
	clog(INFO, "%s: signal pipe: fd %d\n", __func__, signal_pipe[0]);
	para_install_sighandler(SIGINT);
	para_install_sighandler(SIGTERM);
	para_install_sighandler(SIGCHLD);
	para_install_sighandler(SIGHUP);
	signal(SIGPIPE, SIG_IGN);
}

static void audiod_status_dump(void)
{
	static char *prev_status;
	char *tmp = audiod_status_string();

	if (!prev_status || strcmp(tmp, prev_status))
		num_stat_clients = dump_stat_line(tmp);
	free(prev_status);
	prev_status = tmp;
}

static long unsigned penalty_to_ms(int penalty)
{
//	return 1 << (penalty + 6);
	return MAX_PENALTY_MS * penalty * penalty / MAX_PENALTY / MAX_PENALTY;
}

static void clear_slot(int slot)
{
	struct stream_io *s = &siol[slot];

	clog(INFO, "%s: clearing slot %d\n", __func__, slot);
	free(s->buf);
	s->rpid = 0;
	s->wpid = 0;
	s->read_fd = 0;
	s->write_fd = 0;
	s->format = -1;
	s->rkilled = 0;
	s->wkilled = 0;
}

static void kill_stream_writer(int slot)
{
	struct stream_io *s = &siol[slot];

	if (s->format < 0 || s->wkilled || s->wpid <= 0)
		return;
	clog(DEBUG, "stopping %s stream writer (slot %d)\n",
		afi[s->format].name, slot);
	kill(s->wpid, SIGTERM);
	s->wkilled = 1;
	s->buflen = 0;
}

static void kill_stream_reader(int slot)
{
	struct stream_io *s = &siol[slot];

	if (s->format < 0 || s->rkilled || s->rpid <= 0)
		return;
	clog(INFO, "stopping %s stream reader (slot %d)\n",
		afi[s->format].name, slot);
	s->rkilled = 1;
	kill(s->rpid, SIGTERM);
}

/*
 * Send SIGTERM to reader and writer if present
 */
static void kill_decoder(int slot)
{
	kill_stream_writer(slot);
	kill_stream_reader(slot);
}

static void kill_all_decoders(void)
{
	int i;

//	clog(DEBUG, "%s: stopping all decoders\n", __func__);
	for (i = 0; i < MAX_STREAM_SLOTS; i++)
		if (siol[i].format >= 0)
			kill_decoder(i);
}

static void recalc_penalty(int format, int lifetime)
{
	if (lifetime > PENALTY_RESET_TIME) {
		afi[format].penalty = 0;
		return;
	}
	if (lifetime > PENALTY_INCREASE_TIME)
		return;
	if (afi[format].penalty >= MAX_PENALTY)
		return;
	afi[format].penalty++;
	clog(NOTICE, "%s: increasing %s penalty to %lu/%i (%lu/%i ms)\n",
		__func__, afi[format].name, afi[format].penalty,
		MAX_PENALTY,
		penalty_to_ms(afi[format].penalty),
		MAX_PENALTY_MS
	);
}

static void check_sigchld(void)
{
	pid_t pid;
	int i;
reap_next_child:
//	clog(DEBUG, "reaping\n");
	pid = para_reap_child();
	if (pid <= 0)
		return;
	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		struct timeval now;
		struct stream_io *s = &siol[i];
		long lifetime;

		if (s->format < 0)
			continue;
		if (pid != s->rpid && pid != s->wpid)
			continue;
		gettimeofday(&now, NULL);
		if (pid == s->rpid) {
			s->rpid = -1;
			lifetime = now.tv_sec - s->rstime.tv_sec;
			clog(INFO, "%s: %s stream reader in slot %d died "
				"after %li secs\n",
				__func__,
				afi[s->format].name,
				i,
				lifetime);
			recalc_penalty(s->format, lifetime);
			break;
		}
		s->wpid = -1;
		lifetime = now.tv_sec - s->wstime.tv_sec;
		clog(INFO, "%s stream writer in slot %d died "
			"after %li secs\n",
			afi[s->format].name, i, lifetime);
		recalc_penalty(s->format, lifetime);
	}
	goto reap_next_child;
}

static int get_empty_slot(void)
{
	int i;
	struct stream_io *s;

	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		s = &siol[i];
		if (s->format < 0)
			return i;
		if (s->read_fd > 0 || s->write_fd > 0 || s->rpid > 0
			|| s->wpid > 0)
			continue;
		clear_slot(i);
		return i;
	}
	return -1;
}

static int decoder_running(int format)
{
	int i, ret = 0;
	struct stream_io *s;

	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		s = &siol[i];
		if (s->format == format && s->rpid > 0)
			ret |= 1;
		if (s->format == format && s->wpid > 0)
			ret |= 2;
	}
	return ret;
}

static void close_stat_pipe(void)
{
	char *msg;
	if (stat_pipe < 0)
		return;
	clog(NOTICE, "%s: closing status pipe\n", __func__);
	close(stat_pipe);
	kill_all_decoders();
	stat_pipe = -1;
	ring_buffer_free();
	dump_empty_status();
	length_seconds = 0;
	offset_seconds = 0;
	audiod_status_dump();
	playing = 0;
	msg = make_message("%s:no connection to para_server\n",
		status_item_list[SI_STATUS_BAR]);
	ring_buffer_add(msg);
	dump_stat_line(msg);
	free(msg);
}

static void clean_exit(int status)
{
	kill_all_decoders();
	if (socket_name)
		unlink(socket_name);
	if (stat_pipe >= 0)
		close_stat_pipe();
	exit(status);
}

static void start_stream_writer(int slot)
{
	int ret, fds[3] = {1, 0, 0};
	struct stream_io *s = &siol[slot];
	struct audio_format_info *a = &afi[siol[slot].format];

	clog(INFO, "%s: starting stream writer: %s\n", __func__, a->wcmd);
	ret = para_exec_cmdline_pid(&s->wpid, a->wcmd, fds);
	if (ret < 0) {
		clog(ERROR, "%s: exec failed (%d)\n", __func__, ret);
		return;
	}
	s->write_fd = fds[0];
	/* we write to this fd in do_select, so we need non-blocking */
	fcntl(s->write_fd, F_SETFL, O_NONBLOCK);
	gettimeofday(&s->wstime, NULL);
	current_decoder = slot;
	clog(NOTICE, "%s: started %s, pid %d, fd %d, slot %d\n",
		__func__, a->name, s->wpid, s->write_fd, slot);
}

static void start_stream_reader(int format)
{
	struct audio_format_info *a = &afi[format];
	struct stream_io *s;
	int ret, slot;
	int fds[3] = {-1, 1, 0};

	clog(INFO, "%s: starting stream reader: %s\n", __func__, a->rcmd);
	slot = get_empty_slot();
	if (slot < 0) {
		clog(EMERG, "%s: no more empty slots!\n", __func__);
		clean_exit(EXIT_FAILURE);
	}
	s = &siol[slot];
	ret = para_exec_cmdline_pid(&s->rpid, a->rcmd, fds);
	if (ret < 0) {
		clog(ERROR, "%s: exec failed (%d)\n", __func__, ret);
		return;
	}
	s->read_fd = fds[1];
	s->buf = para_malloc(a->bufsize * sizeof(char));
	s->buflen = 0;
	s->format = format;
	s->rkilled = 0;
	s->wkilled = 0;
	gettimeofday(&s->rtime, NULL);
	s->wtime = s->rtime;
	s->rstime = s->rtime;
	a->stime = s->rtime;
	clog(NOTICE, "%s: started %s, pid %d, fd %d slot %d\n",
		__func__, a->name, s->rpid, s->read_fd, slot);
}

static int is_frozen(int format)
{
	long unsigned difftime; /* milliseconds */
	struct timeval now, tmp;
	struct audio_format_info *a = &afi[format];

	gettimeofday(&now, NULL);
	tv_diff(&now, &a->stime, &tmp);
	difftime = tv2ms(&tmp);
	return penalty_to_ms(a->penalty) > difftime? 1 : 0;
}

static void start_current_decoder(void)
{
	int i;
	struct audio_format_info *a;

	if (!af_status)
		return;
	for (i = 0, a = afi; a->name; a++, i++)
		if (!strcmp(af_status, a->name))
                        break;
	if (!a->name)
		return;
	if ((decoder_running(i) & 1) || is_frozen(i))
		return;
	start_stream_reader(i);
}

#define CONVEXITY_COEFF 5


/*
 * print status line if line starts with known command.
 */
static void check_stat_line(char *line)
{
	size_t len, ilen;
	const char *s;
	struct timeval max_deviation = {0, 500 * 1000};


	if (!line)
		return;
	len = strlen(line);

	ring_buffer_add(line);
	num_stat_clients = dump_stat_line(line);

	s = status_item_list[SI_STATUS];
	ilen = strlen(s);
	if (!strncmp(line, s, ilen) && line[ilen] == ':') {
		if (len > ilen + 7 && !strncmp(line + ilen, ":playing", 8))
			playing = 1;
		else
			playing = 0;
		return;
	}
	s = status_item_list[SI_FORMAT];
	ilen = strlen(s);
	if (!strncmp(line, s, ilen) && line[ilen] == ':') {
		free(af_status);
		af_status = para_strdup(line + ilen + 1);
		return;
	}
	s = status_item_list[SI_OFFSET];
	ilen = strlen(s);
	if (!strncmp(line, s, ilen) && line[ilen] == ':') {
		offset_seconds = atoi(line + ilen + 1);
		return;
	}
	s = status_item_list[SI_LENGTH];
	ilen = strlen(s);
	if (!strncmp(line, s, ilen) && line[ilen] == ':') {
		length_seconds = atoi(line + ilen + 1);
		return;
	}
	s = status_item_list[SI_STREAM_START];
	ilen = strlen(s);
	if (!strncmp(line, s, ilen) && line[ilen] == ':') {
		struct timeval tmp;
		if (sscanf(line + ilen + 1, "%lu.%lu", &tmp.tv_sec, &tmp.tv_usec) == 2)
			server_stream_start = tmp;
		return;
	}
	s = status_item_list[SI_CURRENT_TIME];
	ilen = strlen(s);
	if (!strncmp(line, s, ilen) && line[ilen] == ':') {
		struct timeval st, now, tmp, diff;
		static int count;
		int sign;
		if (sscanf(line + ilen + 1, "%lu.%lu", &st.tv_sec, &st.tv_usec) != 2)
			return;
		gettimeofday(&now, NULL);
		sign = tv_diff(&st, &now, &diff);
//		clog(NOTICE, "%s: sign = %i, sa_time_diff_sign = %i\n", __func__,
//			sign, sa_time_diff_sign);
		if (!count) {
			sa_time_diff_sign = sign;
			sa_time_diff = diff;
			count++;
			return;
		}
		if (count > 5) {
			int s = tv_diff(&diff, &sa_time_diff, &tmp);
			if (tv_diff(&max_deviation, &tmp, NULL) < 0)
				clog(WARNING, "%s: time diff jump: %lims\n",
					__func__, s * tv2ms(&tmp));
		}
		count++;
		sa_time_diff_sign = tv_convex_combination(
			sa_time_diff_sign * CONVEXITY_COEFF, &sa_time_diff,
			count > 10? sign : sign * CONVEXITY_COEFF, &diff,
			&tmp);
		sa_time_diff = tmp;
		clog(INFO, "%s: time diff (cur/avg): "
			"%li:%lu/%li:%lu\n", __func__,
			sign * diff.tv_sec, (diff.tv_usec + 500) / 1000,
			sa_time_diff_sign * sa_time_diff.tv_sec,
			(sa_time_diff.tv_usec + 500)/ 1000
		);
	}
}

static void handle_signal(int sig)
{
	switch (sig) {
	case SIGCHLD:
		return check_sigchld();
	case SIGINT:
	case SIGTERM:
	case SIGHUP:
		clog(EMERG, "%s: terminating on signal %d\n", __func__, sig);
		clean_exit(EXIT_FAILURE);
		return;
	}
}

static void check_timeouts(void)
{
	struct timeval now;
	int slot, timeout = conf.stream_timeout_arg;

	gettimeofday(&now, NULL);
	for (slot = 0; slot < MAX_STREAM_SLOTS; slot++) {
		struct stream_io *s = &siol[slot];
		if (s->format < 0)
			continue;
		/* check read time */
		if (s->rpid > 0 && !s->rkilled &&
			now.tv_sec > s->rtime.tv_sec + timeout) {
			clog(DEBUG, "%s input buffer (slot %d) not ready\n",
				afi[s->format].name, slot);
			kill_stream_reader(slot);
		}
		/* check write time */
		if (s->wpid > 0 && !s->wkilled &&
			now.tv_sec > s->wtime.tv_sec + timeout) {
			clog(DEBUG, "%s output buffer (slot %d) not ready\n",
				afi[s->format].name, slot);
			kill_stream_writer(slot);
		}
	}
}

static int check_reader(struct stream_io *s)
{
	return s->read_fd > 0 && s->buflen < afi[s->format].bufsize;
}

static int check_writer(struct stream_io *s)
{
	return s->buflen && s->write_fd >= 0;
}

static void close_unused_stream_fds(int i)
{
	struct stream_io *s = &siol[i];

	if (s->write_fd <= 0)
		goto close_reader;
	if ((s->read_fd < 0 && !s->buflen) || s->wpid <= 0) {
		clog(INFO, "%s: slot %d: closing write fd %d\n",
			__func__, i, s->write_fd);
		close(s->write_fd);
		s->write_fd = -1;
	}
close_reader:
	if (s->read_fd <= 0)
		return;
	if (s->write_fd < 0 || s->rpid < 0) {
		clog(INFO, "%s: slot %d: closing read fd %d\n",
			__func__, i, s->read_fd);
		close(s->read_fd);
		s->read_fd = -1;
		s->buflen = 0;
	}
}

static int set_stream_fds(fd_set *rfds, fd_set *wfds)
{
	int i, max_fileno = -1;
	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		struct stream_io *s = &siol[i];

		close_unused_stream_fds(i);
		if (check_reader(s)) {
			clog(DEBUG, "%s: slot %d: buflen = %d, s->read =%d "
				"FDSET for read\n",
				__func__, i, s->buflen, s->read_fd);
			FD_SET(s->read_fd, rfds);
			max_fileno = MAX(s->read_fd, max_fileno);
		}
		if (s->buflen && !s->wpid)
			start_stream_writer(i);
		if (check_writer(s)) {
			clog(DEBUG, "%s: slot %d: buflen = %d, "
				"s->write = %d DFSET for write\n",
				__func__, i, s->buflen, s->write_fd);
				/*
				 * buf already contains data which was not yet
				 * witten
				 */
			FD_SET(s->write_fd, wfds);
			max_fileno = MAX(s->write_fd, max_fileno);
		}
	}
	return max_fileno;
}

static int write_audio_data(int slot)
{
	struct stream_io *s = &siol[slot];
	int rv;

	clog(DEBUG, "%s: writing %d byte to %s pipe\n", __func__, s->buflen,
		afi[s->format].name);
	rv = write(s->write_fd, s->buf, s->buflen);
	if (rv < 0) {
		clog(errno == EPIPE? INFO : ERROR, 
			"%s: write error in slot %d (fd %d): %s\n",
			 __func__, slot, s->write_fd, strerror(errno));
		kill_decoder(slot);
		s->buflen = 0;
	} else if (rv != s->buflen) {
		clog(DEBUG, "partial %s write (%i/%i) for slot %d\n",
			afi[s->format].name, rv, s->buflen, slot);
		s->buflen -= rv;
		memmove(s->buf, s->buf + rv, s->buflen);
	} else {
		gettimeofday(&s->wtime, NULL);
		s->buflen = 0;
	}
	return rv;
}

static int read_audio_data(int slot)
{
	struct stream_io *s = &siol[slot];
	struct audio_format_info *a = &afi[s->format];
	int ret;

//	clog(DEBUG, "reading at most %d byte from %s pipe (fileno %d)\n",
//			a->bufsize, a->name, file_no);
	ret = read(s->read_fd, s->buf + s->buflen, a->bufsize - s->buflen);
	if (ret > 0)
		s->buflen += ret;
	if (s->wpid < 0)
		s->buflen = 0;
	if (ret < 0) {
		clog(ERROR, "%s: %s read error (%s)\n", __func__,
			a->name, strerror(errno));
		kill_decoder(slot);
		s->buflen = 0;
	} else if (ret > 0)
		gettimeofday(&s->rtime, NULL);
	else {
		clog(INFO, "%s: eof in slot %d, closing read fd %d "
			"(%d bytes left)\n", __func__, slot, s->read_fd,
			s->buflen);
		close(s->read_fd);
		s->read_fd = -1;
	}
	clog(DEBUG, "%s: slot %d: buflen=%d\n", __func__, slot, s->buflen);
	return s->buflen;
}

static void do_stream_io(fd_set *rfds, fd_set *wfds)
{
	int i;

	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		struct stream_io *s = &siol[i];
		if (check_reader(s) && FD_ISSET(s->read_fd, rfds))
			read_audio_data(i);
		if (check_writer(s) && FD_ISSET(s->write_fd, wfds))
			write_audio_data(i);
	}
}
static int init_stream_io(void)
{
	int i, j, ret = -1;
	char *p, *q = NULL;

	for (i = 0; i < conf.stream_write_cmd_given; i++) {
		q = para_strdup(conf.stream_write_cmd_arg[i]);
		clog(DEBUG, "%s: read %s\n", __func__, q);
		p = strchr(q, ':');
		if (!p)
			goto out;
		*p = '\0';
		p++;
		for (j = 0; afi[j].name; j++) {
			if (strcmp(q, afi[j].name))
				continue;
			afi[j].wcmd = conf.stream_write_cmd_arg[i]
				+ (p - q);
			clog(INFO, "%s: sw[%d]: %s\n", __func__, j,
				afi[j].wcmd);
			break;
		}
		free(q);
		q = NULL;
	}
	ret = -2;
	for (i = 0; i < conf.stream_read_cmd_given; i++) {
		q = para_strdup(conf.stream_read_cmd_arg[i]);
		p = strchr(q, ':');
		if (!p)
			goto out;
		*p = '\0';
		p++;
		for (j = 0; afi[j].name; j++) {
			if (strcmp(q, afi[j].name))
				continue;
			afi[j].rcmd = conf.stream_read_cmd_arg[i]
				+ (p - q);
			clog(INFO, "%s: sr[%d]: %s\n", __func__, j,
				afi[j].rcmd);
			break;
		}
		free(q);
		q = NULL;
	}
	ret = 1;
out:
	if (q)
		free(q);
	return ret;
}

static int dump_commands(int fd)
{
	char *buf = para_strdup(""), *tmp = NULL; 
	int i;
	ssize_t ret;

	for (i = 0; cmds[i].name; i++) {
		tmp = make_message("%s%s\t%s\n", buf, cmds[i].name,
			cmds[i].description);
		free(buf);
		buf = tmp;
	}
	ret = write(fd, buf, strlen(buf));
	free(buf);
	return ret;
}

int com_help(int fd, int argc, char **argv)
{
	int i, ret = -1;
	char *buf;
	const char *dflt = "No such command. Available commands:\n";

	if (argc < 2) {
		ret = dump_commands(fd);
		goto out;
	}
	for (i = 0; cmds[i].name; i++) {
		if (strcmp(cmds[i].name, argv[0]))
			continue;
		buf = make_message(
			"NAME:\n~~~~~\n\t%s -- %s\n\n"
			"USAGE:\n~~~~~~\n\tpara_audioc %s\n\n"
			"DESCRIPTION:\n~~~~~~~~~~~~\n%s\n",
			argv[0],
			cmds[i].description,
			cmds[i].usage,
			cmds[i].help
		);
		ret = write(fd, buf, strlen(buf));
		free(buf);
		goto out;
	}
	ret = write(fd, dflt, strlen(dflt));
	if (ret > 0)
		ret = dump_commands(fd);
out:
	close(fd);
	return ret;
}

int com_stat(int fd, int argc, char **argv)
{
	int ret = -1;

	if (num_stat_clients > MAX_STAT_CLIENTS) {
		clog(ERROR, "%s: maximal number of stat clients (%d) exceeded\n",
			__func__, MAX_STAT_CLIENTS);
		goto out;
	}
	ret = ring_buffer_dump(fd);
	if (ret > 0)
		num_stat_clients = stat_client_add(fd);
out:
	if (ret < 0)
		close(fd);
	return ret;
}

int com_term(int fd, int argc, char **argv)
{
	kill_all_decoders();
	clog(NOTICE, "%s: terminating on user request\n", __func__);
	close(fd);
	clean_exit(EXIT_SUCCESS);
	return 42;
}

int com_on(int fd, int argc, char **argv)
{
	audiod_status = AUDIOD_ON;
	close(fd);
	return 1;
}

int com_off(int fd, int argc, char **argv)
{
	audiod_status = AUDIOD_OFF;
	close(fd);
	return 1;
}

int com_sb(int fd, int argc, char **argv)
{
	audiod_status = AUDIOD_STANDBY;
	close(fd);
	return 1;
}

int com_cycle(int fd, int argc, char **argv)
{
	switch (audiod_status) {
		case  AUDIOD_ON:
			return com_sb(fd, argc, argv);
			break;
		case  AUDIOD_OFF:
			return com_on(fd, argc, argv);
			break;
		case  AUDIOD_STANDBY:
			return com_off(fd, argc, argv);
			break;
	}
	close(fd);
	return 1;
}

static int check_perms(struct ucred *c)
{
	int i;

	if (!conf.user_allow_given)
		return 1;
	for (i = 0; i < conf.user_allow_given; i++)
		if (c->uid == conf.user_allow_arg[i])
			return 1;
	return -1;
}

static int handle_connect(void)
{
	int i, argc, ret, clifd = -1;
	size_t len = sizeof(struct sockaddr_un);
	struct ucred c;
	char *buf = para_malloc(MAXLINE), **argv;
	/*
	 * The len argument is a value-result parameter: it should
	 * initially contain the size of the structure pointed  to
	 * by  addr;  on  return it will contain the actual length (in
	 * bytes) of the address returned.
	 */
	ret = -1;
	clifd = accept(audiod_socket, (struct sockaddr *)&unix_addr, &len);
	if (clifd < 0)
		goto out;
	ret = -2;
	len = recv_cred_buffer(clifd, buf, MAXLINE - 1, &c);
	if (len < 0)
		goto out;
	clog(DEBUG, "%s: received: %s (%d bytes)\n", __func__, buf, len);
//	fprintf(stderr, "pid: %i, uid: %i, gid: %i\n", c.pid, c.uid, c.gid);
	ret = -3;
	if (check_perms(&c) <= 0)
		goto out;
	argc = split_args(buf, &argv, '\n');
	clog(DEBUG, "%s: argv[0]: %s\n", __func__, argv[0]);
	for (i = 0; cmds[i].name; i++) {
		if (strcmp(cmds[i].name, argv[0]))
			continue;
		ret = cmds[i].handler(clifd, argc, ++argv);
		goto out;
	}
	ret = -4; /* cmd not found */
	close(clifd);
out:
	free(buf);
	return ret;
}

static void audiod_get_socket(void)
{
	if (conf.socket_given)
		socket_name = para_strdup(conf.socket_arg);
	else {
		char *hn = para_hostname();
		socket_name = make_message("/var/paraslash/audiod_socket.%s",
			hn);
		free(hn);
	}
	clog(NOTICE, "%s: connecting to local socket %s\n", __func__, socket_name);
	if (conf.force_given)
		unlink(socket_name);
	audiod_socket = create_pf_socket(socket_name, &unix_addr,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IWOTH);
	if (audiod_socket < 0) {
		clog(EMERG, "%s: can not connect to socket\n", __func__);
		exit(EXIT_FAILURE); /* do not unlink socket */
	}
	if (listen(audiod_socket, 5) < 0) {
		clog(EMERG, "%s: can not listen on socket\n", __func__);
		exit(EXIT_FAILURE); /* do not unlink socket */
	}
}

static int open_stat_pipe(void)
{
	int ret, fd[3] = {-1, 1, 0};
	char *argv[] = {BINDIR "/para_client", "stat", NULL};
	pid_t pid;
	ret = para_exec(&pid, BINDIR "/para_client",  argv, fd);
	if (ret >= 0) {
		ret = fd[1];
		clog(NOTICE, "%s: stat pipe opened, fd %d\n", __func__, ret);
	}
	return ret;
}

static int audiod_mainloop(void)
{
	fd_set rfds, wfds;
	int ret, max_fileno, sp_numread, sbo = 0;
	char status_buf[STRINGSIZE] = "";
	struct timeval tv;
repeat:
	FD_ZERO(&rfds);
	FD_ZERO(&wfds);
	max_fileno = 0;
	check_timeouts();
	if (audiod_status != AUDIOD_ON)
		kill_all_decoders();
	else if (playing)
		start_current_decoder();
	max_fileno = set_stream_fds(&rfds, &wfds);
	/* stat pipe (read) */
	if (stat_pipe >= 0 && audiod_status == AUDIOD_OFF)
		close_stat_pipe();
	if (stat_pipe < 0 && audiod_status != AUDIOD_OFF)
		stat_pipe = open_stat_pipe();
	if (stat_pipe >= 0 && audiod_status != AUDIOD_OFF) {
		FD_SET(stat_pipe, &rfds);
		max_fileno = MAX(max_fileno, stat_pipe);
	}
	/* signal pipe */
	FD_SET(signal_pipe[0], &rfds);
	max_fileno = MAX(max_fileno, signal_pipe[0]);
	/* local socket */
	if (audiod_socket < 0)
		audiod_get_socket();
	FD_SET(audiod_socket, &rfds);
	max_fileno = MAX(max_fileno, audiod_socket);
	tv.tv_sec = 0;
	tv.tv_usec = 200000;
	ret = select(max_fileno + 1, &rfds, &wfds, NULL, &tv);

	if (ret < 0 && errno != EINTR)
		clog(ERROR, "%s select returned %d (%s)\n", __func__,
			ret, strerror(errno));
	if (audiod_status != AUDIOD_OFF)
		audiod_status_dump();
	if (ret <= 0)
		goto repeat;
	do_stream_io(&rfds, &wfds);
	/* read status pipe */
	if (stat_pipe >=0 && FD_ISSET(stat_pipe, &rfds)) {
		if ((sp_numread = read(stat_pipe, status_buf + sbo,
				STRINGSIZE - 1 - sbo)) <= 0) {
			close_stat_pipe();
			stat_pipe = open_stat_pipe();
			/* avoid busy loop if server is down */
			while (sleep(1) > 0)
				; /* try again*/
		}
		status_buf[sp_numread + sbo] = '\0';
		sbo = check_buf_for_items(status_buf, sp_numread + sbo,
			&check_stat_line, 0);
	}
	if (FD_ISSET(audiod_socket, &rfds)) {
		ret = handle_connect();
		if (ret <= 0)
			clog(ERROR, "%s: command handler returned %d\n", __func__, ret);
	}
	/* signals */
	if (FD_ISSET(signal_pipe[0], &rfds)) {
		int sig_nr = para_next_signal();
		if (sig_nr > 0)
			handle_signal(sig_nr);
	}
	goto repeat;
	return 42; /* never reached */
}

static void set_initial_status(void)
{
	audiod_status = AUDIOD_ON;
	if (!conf.mode_given)
		return;
	if (!strcmp(conf.mode_arg, "sb")) {
		audiod_status = AUDIOD_STANDBY;
		return;
	}
	if (!strcmp(conf.mode_arg, "off")) {
		audiod_status = AUDIOD_OFF;
		return;
	}
	if (strcmp(conf.mode_arg, "on"))
		clog(WARNING, "%s: invalid mode\n", __func__);
}

int main(int argc, char **argv)
{
	char *cf;
	int i;


	hostname = para_hostname();
	cmdline_parser(argc, argv, &conf);
	para_drop_privileges(conf.user_arg);
	cf = configfile_exists();
	if (cf) {
		if (cmdline_parser_configfile(cf, &conf, 0, 0, 0)) {
			fprintf(stderr, "parse error in config file\n");
			exit(EXIT_FAILURE);
		}
	}
	log_welcome("para_audiod", conf.loglevel_arg);
	i = init_stream_io();
	if (i < 0) {
		fprintf(stderr, "invalid stream_%s command\n", i == -1? "read" : "write");
		exit(EXIT_FAILURE);
	}
	server_uptime(SET); /* reset server uptime */
	set_initial_status();
	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		siol[i].buf = NULL;
		clear_slot(i);
	}
	init_ring_buffer();
	setup_signal_handling();
	if (conf.daemon_given)
		daemon_init();
	return audiod_mainloop(); /* never returns */
}
