#include "para.h"

int exec_vp_pid_bg(pid_t *pid, const char *file, char *const args[])
{
	if ((*pid = fork()) < 0)
		exit(EXIT_FAILURE);
	if (!(*pid)) {
		/* child */
		execvp(file, args);
		_exit(EXIT_FAILURE);
	}
	return 1;
}

/*
 * fds must be a pointer to three integers, corresponding to stdin, 
 * stdout and stderr respectively.
 * fd[i] < 0 : leave fd alone
 * fd[i] == 0: dup fd to /dev/null
 * fd[i] > 0 : dup fd and return fileno in fd[i]
 */
int para_exec(pid_t *pid, const char *file, char *const args[], int *fds)
{
	int ret = -1, in[2], out[2], err[2], null = -1; /* ;) */
	
	if (fds[0] > 0 && pipe(in) < 0)
		goto out;
	if (fds[1] > 0 && pipe(out) < 0)
		goto err1;
	if (fds[2] > 0 && pipe(err) < 0)
		goto err2;
	ret = -2;
	if (!fds[0] || !fds[1] || !fds[2])
		null = open("/dev/null", O_RDONLY);
	if ((*pid = fork()) < 0)
		exit(EXIT_FAILURE);
	if (!(*pid)) { /* child */
		int i;
		/* close unneeded fds */
		for (i = 0; i < 256; i++)
			if (i != STDIN_FILENO
				&& i != STDOUT_FILENO
				&& i != STDERR_FILENO
				&& i != in[0]
				&& i != in[1]
				&& i != out[0]
				&& i != out[1]
				&& i != err[0]
				&& i != err[1]
				&& i != null
			)
				close(i);
		if (fds[0] >= 0) {
			close(in[1]);
			if (fds[0]) {
				if (in[0] != STDIN_FILENO)
					dup2(in[0], STDIN_FILENO);
			} else
				dup2(null, STDIN_FILENO);
		}
		if (fds[1] >= 0) {
			close(out[0]);
			if (fds[1]) {
				if (out[1] != STDOUT_FILENO)
					dup2(out[1], STDOUT_FILENO);
			} else
				dup2(null, STDOUT_FILENO);
		}
		if (fds[2] >= 0) {
			close(err[0]);
			if (fds[2]) {
				if (err[1] != STDERR_FILENO)
					dup2(err[1], STDERR_FILENO);
			} else
				dup2(null, STDERR_FILENO);
		}
		if (null >= 0)
			close(null);
		execvp(file, args);
		_exit(EXIT_FAILURE);
	}
	if (fds[0] > 0) {
		close(in[0]);
		*fds = in[1];
	}
	if (fds[1] > 0) {
		close(out[1]);
		*(fds + 1) = out[0];
	}
	if (fds[2] > 0) {
		close(err[1]);
		*(fds + 2) = err[0];
	}
	ret = 1;
		goto out;
	close(err[0]);
	close(err[1]);
err2:	
	close(out[0]);
	close(out[1]);
err1:	
	close(in[0]);
	close(in[1]);
out:
	if (null >= 0)
		close(null);
	return ret;
}


FILE *popen_read_vp_pid(pid_t *pid, const char *file, char *const args[])
{
	//pid_t pid;
	int fd[2];

	if (pipe(fd) < 0)
		return NULL;
	if ((*pid = fork()) < 0)
		exit(EXIT_FAILURE);
	if (!(*pid)) {
		/* child */
		int null = open("/dev/null", O_RDONLY);
		close(fd[0]);
		if (fd[1] != STDOUT_FILENO)
			dup2(fd[1], STDOUT_FILENO);
		dup2(null, STDERR_FILENO);
		close(fd[1]);
		close(null);
		execvp(file, args);
		_exit(EXIT_FAILURE);
	}
	/* parent */
	close(fd[1]);
	return fdopen(fd[0], "r");
}

static void va2array(char** array, char **theargs)
{
	char **argptr = theargs;
	int i = 0;

	while (*argptr && i < 1023) {
	//printf("theargs=%s\n",*argptr);
		array[i] = *argptr;
		argptr++;
		i++;
	}
	array[i] = NULL;
	return;
}

FILE *popen_read_client_pid(pid_t *pid, char **cmd)
{
	char *array[1024];

	array[0] = "client";
	va2array(array + 1, cmd);
	return  popen_read_vp_pid(pid, BINDIR "/para_client", array);
}

FILE *popen_read_client(char **cmd)
{
	pid_t pid;
	return popen_read_client_pid(&pid, cmd);
}

int split_args(char *args, char ***argv_ptr, int delim)
{
	char *p = args;
	char **argv;
	ssize_t n = 0, i;

	while ((p = strchr(p, delim))) {
		p++;
		n++;
	}
	*argv_ptr = malloc((n + 3) * sizeof(char *));
	argv = *argv_ptr;
	i = 0;
	p = args;
//	printf("split_args: a:%s\n", p);
	while (p) {
		argv[i] = p;
	 	p = strchr(p, delim);
		if (p) {
//		printf("a:%s\n", p);
			*p = '\0';
//		printf("b:\n");
			p++;
			i++;
		}
	}
	argv[n + 1] = NULL;
	return n;
}

FILE *popen_read_client_cmdline_pid(pid_t *pid, char *cmdline)
{
	int argc;
	char **argv;

	argc = split_args(cmdline, &argv, ' ');
	if (!argv)
		return NULL;
	return popen_read_client_pid(pid, argv);
}

int para_exec_cmdline_pid(pid_t *pid, char *cmdline, int *fds)
{
	int argc, ret = -1;
	char **argv, *tmp = strdup(cmdline);

	if (!tmp)
		exit(EXIT_FAILURE);
	argc = split_args(tmp, &argv, ' ');
	if (argv)
		ret = para_exec(pid, argv[0], argv, fds);
	free(tmp);
	return ret;
}

FILE *popen_read_client_cmdline(char *cmdline)
{
	pid_t pid;
	
	return popen_read_client_cmdline_pid(&pid, cmdline);
}

FILE *popen_read_cmdline_pid(pid_t *pid, char *cmdline)
{
	int argc;
	char **argv;
	char *tmp = strdup(cmdline);
	FILE *pipe;

	if (!tmp)
		return NULL;
	argc = split_args(tmp, &argv, ' ');
	if (!argv)
		return NULL;
	pipe = popen_read_vp_pid(pid, argv[0], argv);
	free(tmp);
	return pipe;
}

int file_exists(const char *fn)
{
        struct stat statbuf;

        return !stat(fn, &statbuf);
}
int exec_cmdline_pid_bg(pid_t *pid, const char *cmdline)
{
	char *tmp = strdup(cmdline);
	char **argv;
	int ret;

	if (!tmp)
		return -1;
	split_args(tmp, &argv, ' ');
	if (!argv)
		return -1;
	ret = exec_vp_pid_bg(pid, argv[0], argv);
	free(tmp);
	return ret;
}

/*
 * Check if file is fifo, create if neccessary and open with given
 * flags.
 */
int para_open_fifo(const char *fifoname, const int flags)
{
	int ret, fifo;
	struct stat stat_buf;

	ret = stat(fifoname, &stat_buf);
	if (ret < 0) {
		ret = mknod(fifoname, S_IFIFO | S_IRUSR | S_IWUSR, 42);
		if (ret >= 0)
			ret = stat(fifoname, &stat_buf);
		if (ret < 0)
			return ret;
	}
	if (!S_ISFIFO(stat_buf.st_mode))
		return -1;
	fifo = open(fifoname, flags);
	if (fifo < 0)
		return -1;
	return fifo;
}
